@@ -49,7 +49,6 @@ The parts not included are excluded by #ifndef UNIV_INNOCHECKSUM. */
 #include "page0zip.h"            /* page_zip_*() */
 #include "trx0undo.h"            /* TRX_* */
 #include "ut0crc32.h"            /* ut_crc32_init() */
-#include "fsp0pagecompress.h"    /* fil_get_compression_alg_name */
 #include "fil0crypt.h"           /* fil_space_verify_crypt_checksum */
 
 #include <string.h>
@@ -104,6 +103,9 @@ FILE*				log_file = NULL;
 /* Enabled for log write option. */
 static bool			is_log_enabled = false;
 
+static byte field_ref_zero_buf[UNIV_PAGE_SIZE_MAX];
+const byte *field_ref_zero = field_ref_zero_buf;
+
 #ifndef _WIN32
 /* advisory lock for non-window system. */
 struct flock			lk;
@@ -471,7 +473,7 @@ is_page_corrupted(
 	/* use to store LSN values. */
 	uint32_t logseq;
 	uint32_t logseqfield;
-	ulint page_type = mach_read_from_2(buf+FIL_PAGE_TYPE);
+	const uint16_t page_type = mach_read_from_2(buf+FIL_PAGE_TYPE);
 	uint32_t key_version = buf_page_get_key_version(buf, flags);
 	uint32_t space_id = mach_read_from_4(
 		buf + FIL_PAGE_ARCH_LOG_NO_OR_SPACE_ID);
@@ -816,7 +818,7 @@ static inline bool is_page_free(const byte *xdes, ulint physical_page_size,
   const byte *des=
       xdes + XDES_ARR_OFFSET +
       XDES_SIZE * ((page_no & (physical_page_size - 1)) / FSP_EXTENT_SIZE);
-  return xdes_get_bit(des, XDES_FREE_BIT, page_no % FSP_EXTENT_SIZE);
+  return xdes_is_free(des, page_no % FSP_EXTENT_SIZE);
 }
 
 /*
@@ -845,7 +847,7 @@ parse_page(
 	/* Check whether page is doublewrite buffer. */
 	str = skip_page ? "Double_write_buffer" : "-";
 
-	switch (mach_read_from_2(page + FIL_PAGE_TYPE)) {
+	switch (fil_page_get_type(page)) {
 
 	case FIL_PAGE_INDEX: {
 		uint32_t key_version = mach_read_from_4(page + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
@@ -896,8 +898,7 @@ parse_page(
 			/* update per-index statistics */
 			{
 				per_index_stats &index = index_ids[id];
-				if (is_page_free(xdes, physical_page_size,
-						 page_no)) {
+				if (is_page_free(xdes, physical_page_size, page_no)) {
 					index.free_pages++;
 					return;
 				}
@@ -1192,15 +1193,11 @@ print_summary(
 	fprintf(fil_out, "index_id\t#pages\t\t#leaf_pages\t#recs_per_page"
 		"\t#bytes_per_page\n");
 
-	for (const auto &ids : index_ids) {
-		const per_index_stats& index = ids.second;
-		if (!index.pages) {
-			DBUG_ASSERT(index.free_pages);
-			continue;
-		}
-
+	for (std::map<unsigned long long, per_index_stats>::const_iterator it = index_ids.begin();
+	     it != index_ids.end(); it++) {
+		const per_index_stats& index = it->second;
 		fprintf(fil_out, "%lld\t\t%lld\t\t%lld\t\t%lld\t\t%lld\n",
-			ids.first, index.pages, index.leaf_pages,
+			it->first, index.pages, index.leaf_pages,
 			index.total_n_recs / index.pages,
 			index.total_data_bytes / index.pages);
 	}
@@ -1309,11 +1306,11 @@ static void usage(void)
 
 extern "C" my_bool
 innochecksum_get_one_option(
-	int			optid,
-	const struct my_option	*opt MY_ATTRIBUTE((unused)),
-	char			*argument MY_ATTRIBUTE((unused)))
+	const struct my_option	*opt,
+	const char		*argument MY_ATTRIBUTE((unused)),
+        const char *)
 {
-	switch (optid) {
+	switch (opt->id) {
 #ifndef DBUG_OFF
 	case '#':
 		dbug_setting = argument
@@ -1528,8 +1525,6 @@ int main(
 	/* our input filename. */
 	char*		filename;
 	/* Buffer to store pages read. */
-	byte*		buf_ptr = NULL;
-	byte*		xdes_ptr = NULL;
 	byte*		buf = NULL;
 	byte*		xdes = NULL;
 	/* bytes read count */
@@ -1563,7 +1558,6 @@ int main(
 	/* enable when space_id of given file is zero. */
 	bool		is_system_tablespace = false;
 
-	ut_crc32_init();
 	MY_INIT(argv[0]);
 	DBUG_ENTER("main");
 	DBUG_PROCESS(argv[0]);
@@ -1609,10 +1603,10 @@ int main(
 	}
 
 
-	buf_ptr = (byte*) malloc(UNIV_PAGE_SIZE_MAX * 2);
-	xdes_ptr = (byte*)malloc(UNIV_PAGE_SIZE_MAX * 2);
-	buf = (byte *) ut_align(buf_ptr, UNIV_PAGE_SIZE_MAX);
-	xdes = (byte *) ut_align(xdes_ptr, UNIV_PAGE_SIZE_MAX);
+	buf = static_cast<byte*>(aligned_malloc(UNIV_PAGE_SIZE_MAX,
+						UNIV_PAGE_SIZE_MAX));
+	xdes = static_cast<byte*>(aligned_malloc(UNIV_PAGE_SIZE_MAX,
+						 UNIV_PAGE_SIZE_MAX));
 
 	/* The file name is not optional. */
 	for (int i = 0; i < argc; ++i) {
@@ -1895,6 +1889,18 @@ unexpected_eof:
 			}
 
 			if (ferror(fil_in)) {
+#ifdef _AIX
+				/*
+				  AIX fseeko can go past eof without error.
+				  the error occurs on read, hence output the
+				  same error here as would show up on other
+				  platforms. This shows up in the mtr test
+				  innodb_zip.innochecksum_3-4k,crc32,innodb
+				*/
+				if (errno == EFBIG) {
+					goto unexpected_eof;
+				}
+#endif
 				fprintf(stderr, "Error reading " ULINTPF " bytes",
 					physical_page_size);
 				perror(" ");
@@ -1919,7 +1925,7 @@ first_non_zero:
 				skip_page = false;
 			}
 
-			ulint cur_page_type = mach_read_from_2(buf+FIL_PAGE_TYPE);
+			const uint16_t cur_page_type = fil_page_get_type(buf);
 
 			/* FIXME: Page compressed or Page compressed and encrypted
 			pages do not contain checksum. */
@@ -1932,8 +1938,7 @@ first_non_zero:
 			checksum verification.*/
 			if (!no_check
 			    && !skip_page
-			    && !is_page_free(xdes, physical_page_size,
-					     cur_page_num)
+			    && !is_page_free(xdes, physical_page_size, cur_page_num)
 			    && (exit_status = verify_checksum(
 						buf, is_encrypted,
 						&mismatch_count, flags))) {
@@ -2003,21 +2008,9 @@ first_non_zero:
 		fclose(log_file);
 	}
 
-	free(buf_ptr);
-	free(xdes_ptr);
-
-	my_end(exit_status);
-	DBUG_RETURN(exit_status);
+	goto common_exit;
 
 my_exit:
-	if (buf_ptr) {
-		free(buf_ptr);
-	}
-
-	if (xdes_ptr) {
-		free(xdes_ptr);
-	}
-
 	if (!read_from_stdin && fil_in) {
 		fclose(fil_in);
 	}
@@ -2026,6 +2019,9 @@ my_exit:
 		fclose(log_file);
 	}
 
+common_exit:
+	aligned_free(buf);
+	aligned_free(xdes);
 	my_end(exit_status);
 	DBUG_RETURN(exit_status);
 }
