@@ -27,9 +27,6 @@ struct rpl_parallel_thread_pool global_rpl_thread_pool;
 
 static void signal_error_to_sql_driver_thread(THD *thd, rpl_group_info *rgi,
                                               int err);
-static void
-register_wait_for_prior_event_group_commit(rpl_group_info *rgi,
-                                           rpl_parallel_entry *entry);
 
 static int
 rpt_handle_event(rpl_parallel_thread::queued_event *qev,
@@ -154,35 +151,15 @@ finish_event_group(rpl_parallel_thread *rpt, uint64 sub_id,
   int err;
 
   thd->get_stmt_da()->set_overwrite_status(true);
-
-  if (unlikely(rgi->worker_error))
-  {
-    /*
-      In case a previous wait was killed, we need to re-register to be able to
-      repeat the wait.
-
-      And before doing that, we un-register any previous registration (in case
-      we got an error earlier and skipped waiting).
-    */
-    thd->wait_for_commit_ptr->unregister_wait_for_prior_commit();
-    mysql_mutex_lock(&entry->LOCK_parallel_entry);
-    register_wait_for_prior_event_group_commit(rgi, entry);
-    mysql_mutex_unlock(&entry->LOCK_parallel_entry);
-  }
-
   /*
     Remove any left-over registration to wait for a prior commit to
     complete. Normally, such wait would already have been removed at
     this point by wait_for_prior_commit() called from within COMMIT
-    processing.
-
-    However, in case of MyISAM and no binlog, we might not have any commit
-    processing, and so we need to do the wait here, before waking up any
-    subsequent commits, to preserve correct order of event execution.
-
-    Also, in the error case we might have skipped waiting and thus need to
-    remove it explicitly. Or the wait might have been killed and we need to
-    repeat the registration and the wait.
+    processing. However, in case of MyISAM and no binlog, we might not
+    have any commit processing, and so we need to do the wait here,
+    before waking up any subsequent commits, to preserve correct
+    order of event execution. Also, in the error case we might have
+    skipped waiting and thus need to remove it explicitly.
 
     It is important in the non-error case to do a wait, not just an
     unregister. Because we might be last in a group-commit that is
@@ -195,18 +172,8 @@ finish_event_group(rpl_parallel_thread *rpt, uint64 sub_id,
     all earlier event groups have also committed; this way no more
     mark_start_commit() calls can be made and it is safe to de-allocate
     the GCO.
-
-    Thus this final wait is done with kill ignored during the wait. This is
-    fine, at this point there is no active query or transaction to abort, and
-    the thread will continue as soon as earlier event groups complete.
-
-    Note though, that in the non-error case there is no guarantee that
-    finish_event_group() will be run in-order. For example, a successful
-    binlog group commit will wakeup all participating event groups
-    simultaneously so only thread scheduling will decide the order in which
-    finish_event_group() calls acquire LOCK_parallel_entry.
   */
-  err= wfc->wait_for_prior_commit(thd, false);
+  err= wfc->wait_for_prior_commit(thd);
   if (unlikely(err) && !rgi->worker_error)
     signal_error_to_sql_driver_thread(thd, rgi, err);
   thd->wait_for_commit_ptr= NULL;
@@ -218,7 +185,6 @@ finish_event_group(rpl_parallel_thread *rpt, uint64 sub_id,
     waiting for this). In most cases (normal DML), it will be a no-op.
   */
   rgi->mark_start_commit_no_lock();
-  rgi->commit_orderer.wakeup_blocked= false;
 
   if (entry->last_committed_sub_id < sub_id)
   {
@@ -276,7 +242,8 @@ finish_event_group(rpl_parallel_thread *rpt, uint64 sub_id,
     not yet started should just skip their group, preparing for stop of the
     SQL driver thread.
   */
-  if (unlikely(rgi->worker_error) && entry->stop_on_error_sub_id > sub_id)
+  if (unlikely(rgi->worker_error) &&
+      entry->stop_on_error_sub_id == (uint64)ULONGLONG_MAX)
     entry->stop_on_error_sub_id= sub_id;
   mysql_mutex_unlock(&entry->LOCK_parallel_entry);
 #ifdef ENABLED_DEBUG_SYNC
@@ -294,12 +261,6 @@ finish_event_group(rpl_parallel_thread *rpt, uint64 sub_id,
                               STRING_WITH_LEN("now WAIT_FOR proceed_by_1000"));
       }
     });
-  DBUG_EXECUTE_IF("hold_worker2_favor_worker3", {
-      if (rgi->current_gtid.seq_no == 2001) {
-        DBUG_ASSERT(!rgi->worker_error || entry->stop_on_error_sub_id == sub_id);
-        debug_sync_set_action(thd, STRING_WITH_LEN("now SIGNAL cont_worker3"));
-      }
-    });
 #endif
 
   if (rgi->killed_for_retry == rpl_group_info::RETRY_KILL_PENDING)
@@ -319,11 +280,11 @@ static void
 signal_error_to_sql_driver_thread(THD *thd, rpl_group_info *rgi, int err)
 {
   rgi->worker_error= err;
-  DBUG_EXECUTE_IF("hold_worker2_favor_worker3", {
-      if (rgi->current_gtid.seq_no == 2002) {
-        debug_sync_set_action(thd, STRING_WITH_LEN("now WAIT_FOR cont_worker2"));
-      }});
-
+  /*
+    In case we get an error during commit, inform following transactions that
+    we aborted our commit.
+  */
+  rgi->unmark_start_commit();
   rgi->cleanup_context(thd, true);
   rgi->rli->abort_slave= true;
   rgi->rli->stop_for_until= false;
@@ -370,7 +331,7 @@ register_wait_for_prior_event_group_commit(rpl_group_info *rgi,
   Do not start parallel execution of this event group until all prior groups
   have reached the commit phase that are not safe to run in parallel with.
 */
-static void
+static bool
 do_gco_wait(rpl_group_info *rgi, group_commit_orderer *gco,
             bool *did_enter_cond, PSI_stage_info *old_stage)
 {
@@ -422,45 +383,18 @@ do_gco_wait(rpl_group_info *rgi, group_commit_orderer *gco,
                       &entry->LOCK_parallel_entry);
     } while (wait_count > entry->count_committing_event_groups);
   }
-}
 
-
-static bool
-do_stop_handling(rpl_group_info *rgi)
-{
-  bool should_stop= false;
-  rpl_parallel_entry *entry=  rgi->parallel_entry;
-
-  mysql_mutex_assert_owner(&entry->LOCK_parallel_entry);
-
-  if (unlikely(entry->force_abort) && rgi->gtid_sub_id > entry->stop_sub_id)
+  if (entry->force_abort && wait_count > entry->stop_count)
   {
     /*
-      We are stopping (STOP SLAVE), and this event group need not be applied
-      before we can safely stop. So return a flag that will cause us to skip,
-      rather than execute, the following events. Once all queued events have
-      been skipped, the STOP SLAVE is complete (for this thread).
+      We are stopping (STOP SLAVE), and this event group is beyond the point
+      where we can safely stop. So return a flag that will cause us to skip,
+      rather than execute, the following events.
     */
-    should_stop= true;
+    return true;
   }
-
-  if (unlikely(entry->stop_on_error_sub_id <= rgi->wait_commit_sub_id))
-  {
-    rgi->worker_error= 1;
-    should_stop= true;
-  }
-
-  if (likely(!should_stop))
-  {
-    /*
-      Since we did not decide to stop, bump the largest_started_sub_id while
-      still holding LOCK_parallel_entry.
-    */
-    if (rgi->gtid_sub_id > entry->largest_started_sub_id)
-      entry->largest_started_sub_id= rgi->gtid_sub_id;
-  }
-
-  return should_stop;
+  else
+    return false;
 }
 
 
@@ -507,25 +441,15 @@ do_ftwrl_wait(rpl_group_info *rgi,
       mysql_cond_wait(&entry->COND_parallel_entry, &entry->LOCK_parallel_entry);
     } while (sub_id > entry->pause_sub_id);
 
-    DBUG_EXECUTE_IF("delay_ftwrl_wait_gtid_0_x_100", {
-        if (rgi->current_gtid.domain_id == 0 &&
-            rgi->current_gtid.seq_no == 100) {
-          /*
-            Simulate delayed wakeup from the mysql_cond_wait(). To do this, we
-            need to have the LOCK_parallel_entry mutex released during the wait.
-          */
-          mysql_mutex_unlock(&entry->LOCK_parallel_entry);
-          debug_sync_set_action(thd,
-              STRING_WITH_LEN("now SIGNAL pause_wait_started WAIT_FOR pause_wait_continue"));
-          mysql_mutex_lock(&entry->LOCK_parallel_entry);
-        }
-      });
     /*
       We do not call EXIT_COND() here, as this will be done later by our
       caller (since we set *did_enter_cond to true).
     */
   }
 
+  if (sub_id > entry->largest_started_sub_id)
+    entry->largest_started_sub_id= sub_id;
+
   DBUG_RETURN(aborted);
 }
 
@@ -683,17 +607,7 @@ rpl_pause_for_ftwrl(THD *thd)
     mysql_mutex_unlock(&rpt->LOCK_rpl_thread);
     ++e->need_sub_id_signal;
     if (e->pause_sub_id == (uint64)ULONGLONG_MAX)
-    {
       e->pause_sub_id= e->largest_started_sub_id;
-      DBUG_EXECUTE_IF("pause_for_ftwrl_wait", {
-          mysql_mutex_unlock(&e->LOCK_parallel_entry);
-          debug_sync_set_action(thd,
-                                STRING_WITH_LEN("now "
-                                                "SIGNAL pause_ftwrl_waiting "
-                                                "WAIT_FOR pause_ftwrl_cont"));
-          mysql_mutex_lock(&e->LOCK_parallel_entry);
-        });
-    }
     thd->ENTER_COND(&e->COND_parallel_entry, &e->LOCK_parallel_entry,
                     &stage_waiting_for_ftwrl_threads_to_pause, &old_stage);
     thd->set_time_for_next_stage();
@@ -780,12 +694,14 @@ convert_kill_to_deadlock_error(rpl_group_info *rgi)
 static int
 is_group_ending(Log_event *ev, Log_event_type event_type)
 {
-  if (event_type == XID_EVENT)
+  if (event_type == XID_EVENT || event_type == XA_PREPARE_LOG_EVENT)
     return 1;
   if (event_type == QUERY_EVENT)  // COMMIT/ROLLBACK are never compressed
   {
     Query_log_event *qev = (Query_log_event *)ev;
-    if (qev->is_commit())
+    if (qev->is_commit() ||
+        !strncmp(qev->query, STRING_WITH_LEN("XA COMMIT")) ||
+        !strncmp(qev->query, STRING_WITH_LEN("XA ROLLBACK")))
       return 1;
     if (qev->is_rollback())
       return 2;
@@ -856,7 +772,8 @@ do_retry:
     thd->wait_for_commit_ptr->unregister_wait_for_prior_commit();
   DBUG_EXECUTE_IF("inject_mdev8031", {
       /* Simulate that we get deadlock killed at this exact point. */
-      slave_background_kill_request(thd);
+      rgi->killed_for_retry= rpl_group_info::RETRY_KILL_KILLED;
+      thd->set_killed(KILL_CONNECTION);
   });
 #ifdef ENABLED_DEBUG_SYNC
   DBUG_EXECUTE_IF("rpl_parallel_simulate_wait_at_retry", {
@@ -873,14 +790,7 @@ do_retry:
   thd->reset_killed();
   thd->clear_error();
   rgi->killed_for_retry = rpl_group_info::RETRY_KILL_NONE;
-#ifdef ENABLED_DEBUG_SYNC
-    DBUG_EXECUTE_IF("hold_worker2_favor_worker3", {
-      if (rgi->current_gtid.seq_no == 2003) {
-        debug_sync_set_action(thd,
-                              STRING_WITH_LEN("now WAIT_FOR cont_worker3"));
-      }
-    });
-#endif
+
   /*
     If we retry due to a deadlock kill that occurred during the commit step, we
     might have already updated (but not committed) an update of table
@@ -898,18 +808,18 @@ do_retry:
   for (;;)
   {
     mysql_mutex_lock(&entry->LOCK_parallel_entry);
-    if (rgi->gtid_sub_id < entry->stop_on_error_sub_id
+    if (entry->stop_on_error_sub_id == (uint64) ULONGLONG_MAX ||
 #ifndef DBUG_OFF
-        || DBUG_EVALUATE_IF("simulate_mdev_12746", 1, 0)
+        (DBUG_EVALUATE_IF("simulate_mdev_12746", 1, 0)) ||
 #endif
-        )
+        rgi->gtid_sub_id < entry->stop_on_error_sub_id)
     {
       register_wait_for_prior_event_group_commit(rgi, entry);
     }
     else
     {
       /*
-        A failure of a preceeding "parent" transaction may not be
+        A failure of a preceding "parent" transaction may not be
         seen by the current one through its own worker_error.
         Such induced error gets set by ourselves now.
       */
@@ -1284,11 +1194,6 @@ handle_rpl_parallel_thread(void *arg)
         bool did_enter_cond= false;
         PSI_stage_info old_stage;
 
-        DBUG_EXECUTE_IF("rpl_parallel_delay_gtid_0_x_100_start", {
-            if (rgi->current_gtid.domain_id==0 &&
-                rgi->current_gtid.seq_no == 100)
-              my_sleep(10000);
-          });
 #ifdef ENABLED_DEBUG_SYNC
         DBUG_EXECUTE_IF("hold_worker_on_schedule", {
             if (rgi->current_gtid.domain_id == 0 &&
@@ -1335,15 +1240,14 @@ handle_rpl_parallel_thread(void *arg)
         event_gtid_sub_id= rgi->gtid_sub_id;
         rgi->thd= thd;
 
-        DBUG_EXECUTE_IF("gco_wait_delay_gtid_0_x_99", {
-            if (rgi->current_gtid.domain_id == 0 && rgi->current_gtid.seq_no == 99) {
-              debug_sync_set_action(thd,
-                  STRING_WITH_LEN("now SIGNAL gco_wait_paused WAIT_FOR gco_wait_cont"));
-            } });
-
         mysql_mutex_lock(&entry->LOCK_parallel_entry);
-        do_gco_wait(rgi, gco, &did_enter_cond, &old_stage);
-        skip_event_group= do_stop_handling(rgi);
+        skip_event_group= do_gco_wait(rgi, gco, &did_enter_cond, &old_stage);
+
+        if (unlikely(entry->stop_on_error_sub_id <= rgi->wait_commit_sub_id))
+        {
+          skip_event_group= true;
+          rgi->worker_error= 1;
+        }
         if (likely(!skip_event_group))
           skip_event_group= do_ftwrl_wait(rgi, &did_enter_cond, &old_stage);
 
@@ -1425,22 +1329,7 @@ handle_rpl_parallel_thread(void *arg)
         if (!thd->killed)
         {
           DEBUG_SYNC(thd, "rpl_parallel_before_mark_start_commit");
-          if (thd->lex->stmt_accessed_temp_table())
-          {
-            /*
-              Temporary tables are special, they require strict
-              single-threaded use as they have no locks protecting concurrent
-              access. Therefore, we cannot safely use the optimization of
-              overlapping the commit of this transaction with the start of the
-              following.
-              So we skip the early mark_start_commit() and also block any
-              wakeup_subsequent_commits() until this event group is fully
-              done, inside finish_event_group().
-            */
-            rgi->commit_orderer.wakeup_blocked= true;
-          }
-          else
-            rgi->mark_start_commit();
+          rgi->mark_start_commit();
           DEBUG_SYNC(thd, "rpl_parallel_after_mark_start_commit");
         }
       }
@@ -1483,13 +1372,8 @@ handle_rpl_parallel_thread(void *arg)
                         err= dbug_simulate_tmp_error(rgi, thd););
         if (unlikely(err))
         {
-          ulong max_retries= slave_trans_retries;
           convert_kill_to_deadlock_error(rgi);
-          DBUG_EXECUTE_IF("rpl_mdev31655_zero_retries",
-                          if ((rgi->current_gtid.seq_no % 1000) == 0)
-                            max_retries= 0;
-                          );
-          if (has_temporary_error(thd) && max_retries > 0)
+          if (has_temporary_error(thd) && slave_trans_retries > 0)
             err= retry_event_group(rgi, rpt, qev);
         }
       }
@@ -1701,7 +1585,7 @@ rpl_parallel_change_thread_count(rpl_parallel_thread_pool *pool,
     to allocate, and will not be left with a half-functional thread pool.
   */
   if (new_count &&
-      !my_multi_malloc(MYF(MY_WME|MY_ZEROFILL),
+      !my_multi_malloc(PSI_INSTRUMENT_ME, MYF(MY_WME|MY_ZEROFILL),
                        &new_list, new_count*sizeof(*new_list),
                        &rpt_array, new_count*sizeof(*rpt_array),
                        NULL))
@@ -1933,7 +1817,7 @@ rpl_parallel_thread::get_qev_common(Log_event *ev, ulonglong event_size)
   mysql_mutex_assert_owner(&LOCK_rpl_thread);
   if ((qev= qev_free_list))
     qev_free_list= qev->next;
-  else if(!(qev= (queued_event *)my_malloc(sizeof(*qev), MYF(0))))
+  else if(!(qev= (queued_event *)my_malloc(PSI_INSTRUMENT_ME, sizeof(*qev), MYF(0))))
   {
     my_error(ER_OUTOFMEMORY, MYF(0), (int)sizeof(*qev));
     return NULL;
@@ -2096,7 +1980,7 @@ rpl_parallel_thread::get_gco(uint64 wait_count, group_commit_orderer *prev,
   mysql_mutex_assert_owner(&LOCK_rpl_thread);
   if ((gco= gco_free_list))
     gco_free_list= gco->next_gco;
-  else if(!(gco= (group_commit_orderer *)my_malloc(sizeof(*gco), MYF(0))))
+  else if(!(gco= (group_commit_orderer *)my_malloc(PSI_INSTRUMENT_ME, sizeof(*gco), MYF(0))))
   {
     my_error(ER_OUTOFMEMORY, MYF(0), (int)sizeof(*gco));
     return NULL;
@@ -2109,9 +1993,6 @@ rpl_parallel_thread::get_gco(uint64 wait_count, group_commit_orderer *prev,
   gco->prior_sub_id= prior_sub_id;
   gco->installed= false;
   gco->flags= 0;
-#ifndef DBUG_OFF
-  gco->gc_done= false;
-#endif
   return gco;
 }
 
@@ -2119,10 +2000,6 @@ rpl_parallel_thread::get_gco(uint64 wait_count, group_commit_orderer *prev,
 void
 rpl_parallel_thread::loc_free_gco(group_commit_orderer *gco)
 {
-#ifndef DBUG_OFF
-  DBUG_ASSERT(!gco->gc_done);
-  gco->gc_done= true;
-#endif
   if (!loc_gco_list)
     loc_gco_last_ptr_ptr= &gco->next_gco;
   else
@@ -2253,23 +2130,34 @@ rpl_parallel_thread_pool::release_thread(rpl_parallel_thread *rpt)
   and the LOCK_rpl_thread must be released with THD::EXIT_COND() instead
   of mysql_mutex_unlock.
 
-  If the flag `reuse' is set, the last worker thread will be returned again,
+  When `gtid_ev' is not NULL the last worker thread will be returned again,
   if it is still available. Otherwise a new worker thread is allocated.
+
+  A worker for XA transaction is determined through xid hashing which
+  ensure for a XA-complete to be scheduled to the same-xid XA-prepare worker.
 */
 rpl_parallel_thread *
 rpl_parallel_entry::choose_thread(rpl_group_info *rgi, bool *did_enter_cond,
-                                  PSI_stage_info *old_stage, bool reuse)
+                                  PSI_stage_info *old_stage,
+                                  Gtid_log_event *gtid_ev)
 {
   uint32 idx;
   Relay_log_info *rli= rgi->rli;
   rpl_parallel_thread *thr;
 
   idx= rpl_thread_idx;
-  if (!reuse)
+  if (gtid_ev)
   {
-    ++idx;
-    if (idx >= rpl_thread_max)
-      idx= 0;
+    if (gtid_ev->flags2 &
+        (Gtid_log_event::FL_COMPLETED_XA | Gtid_log_event::FL_PREPARED_XA))
+      idx= my_hash_sort(&my_charset_bin, gtid_ev->xid.key(),
+                        gtid_ev->xid.key_length()) % rpl_thread_max;
+    else
+    {
+      ++idx;
+      if (idx >= rpl_thread_max)
+        idx= 0;
+    }
     rpl_thread_idx= idx;
   }
   thr= rpl_threads[idx];
@@ -2296,6 +2184,21 @@ rpl_parallel_entry::choose_thread(rpl_group_info *rgi, bool *did_enter_cond,
         /* The thread is ready to queue into. */
         break;
       }
+      else if (unlikely(rli->sql_driver_thd->check_killed(1)))
+      {
+        unlock_or_exit_cond(rli->sql_driver_thd, &thr->LOCK_rpl_thread,
+                            did_enter_cond, old_stage);
+        my_error(ER_CONNECTION_KILLED, MYF(0));
+#ifdef ENABLED_DEBUG_SYNC
+        DBUG_EXECUTE_IF("rpl_parallel_wait_queue_max",
+          {
+            debug_sync_set_action(rli->sql_driver_thd,
+                      STRING_WITH_LEN("now SIGNAL wait_queue_killed"));
+          };);
+#endif
+        slave_output_error_info(rgi, rli->sql_driver_thd);
+        return NULL;
+      }
       else
       {
         /*
@@ -2323,23 +2226,6 @@ rpl_parallel_entry::choose_thread(rpl_group_info *rgi, bool *did_enter_cond,
                                           old_stage);
           *did_enter_cond= true;
         }
-
-        if (unlikely(rli->sql_driver_thd->check_killed(1)))
-        {
-          unlock_or_exit_cond(rli->sql_driver_thd, &thr->LOCK_rpl_thread,
-                              did_enter_cond, old_stage);
-          my_error(ER_CONNECTION_KILLED, MYF(0));
-#ifdef ENABLED_DEBUG_SYNC
-          DBUG_EXECUTE_IF("rpl_parallel_wait_queue_max",
-            {
-              debug_sync_set_action(rli->sql_driver_thd,
-                        STRING_WITH_LEN("now SIGNAL wait_queue_killed"));
-            };);
-#endif
-          slave_output_error_info(rgi, rli->sql_driver_thd);
-          return NULL;
-        }
-
         mysql_cond_wait(&thr->COND_rpl_thread_queue, &thr->LOCK_rpl_thread);
       }
     }
@@ -2370,7 +2256,7 @@ free_rpl_parallel_entry(void *element)
 rpl_parallel::rpl_parallel() :
   current(NULL), sql_thread_stopping(false)
 {
-  my_hash_init(&domain_hash, &my_charset_bin, 32,
+  my_hash_init(PSI_INSTRUMENT_ME, &domain_hash, &my_charset_bin, 32,
                offsetof(rpl_parallel_entry, domain_id), sizeof(uint32),
                NULL, free_rpl_parallel_entry, HASH_UNIQUE);
 }
@@ -2404,7 +2290,7 @@ rpl_parallel::find(uint32 domain_id)
     if (count == 0 || count > opt_slave_parallel_threads)
       count= opt_slave_parallel_threads;
     rpl_parallel_thread **p;
-    if (!my_multi_malloc(MYF(MY_WME|MY_ZEROFILL),
+    if (!my_multi_malloc(PSI_INSTRUMENT_ME, MYF(MY_WME|MY_ZEROFILL),
                          &e, sizeof(*e),
                          &p, count*sizeof(*p),
                          NULL))
@@ -2417,14 +2303,14 @@ rpl_parallel::find(uint32 domain_id)
     e->domain_id= domain_id;
     e->stop_on_error_sub_id= (uint64)ULONGLONG_MAX;
     e->pause_sub_id= (uint64)ULONGLONG_MAX;
+    if (my_hash_insert(&domain_hash, (uchar *)e))
+    {
+      my_free(e);
+      return NULL;
+    }
     mysql_mutex_init(key_LOCK_parallel_entry, &e->LOCK_parallel_entry,
                      MY_MUTEX_INIT_FAST);
     mysql_cond_init(key_COND_parallel_entry, &e->COND_parallel_entry, NULL);
-    if (my_hash_insert(&domain_hash, (uchar *)e))
-    {
-      free_rpl_parallel_entry(e);
-      return NULL;
-    }
   }
   else
     e->force_abort= false;
@@ -2462,18 +2348,20 @@ rpl_parallel::wait_for_done(THD *thd, Relay_log_info *rli)
       are also executed, so that we stop at a consistent point in the binlog
       stream (per replication domain).
 
-      At this point, we are holding LOCK_parallel_entry, and we know that no
-      event group after e->largest_started_sub_id has started running yet. We
-      record this value in e->stop_sub_id, and then each event group can check
-      their own sub_id against it. If their sub_id is strictly larger, then
-      that event group will be skipped.
+      All event groups wait for e->count_committing_event_groups to reach
+      the value of group_commit_orderer::wait_count before starting to
+      execute. Thus, at this point we know that any event group with a
+      strictly larger wait_count are safe to skip, none of them can have
+      started executing yet. So we set e->stop_count here and use it to
+      decide in the worker threads whether to continue executing an event
+      group or whether to skip it, when force_abort is set.
 
       If we stop due to reaching the START SLAVE UNTIL condition, then we
       need to continue executing any queued events up to that point.
     */
     e->force_abort= true;
-    e->stop_sub_id= rli->stop_for_until ?
-      e->current_sub_id : e->largest_started_sub_id;
+    e->stop_count= rli->stop_for_until ?
+      e->count_queued_event_groups : e->count_committing_event_groups;
     mysql_mutex_unlock(&e->LOCK_parallel_entry);
     for (j= 0; j < e->rpl_thread_max; ++j)
     {
@@ -2529,19 +2417,30 @@ rpl_parallel::stop_during_until()
     e= (struct rpl_parallel_entry *)my_hash_element(&domain_hash, i);
     mysql_mutex_lock(&e->LOCK_parallel_entry);
     if (e->force_abort)
-      e->stop_sub_id= e->largest_started_sub_id;
+      e->stop_count= e->count_committing_event_groups;
     mysql_mutex_unlock(&e->LOCK_parallel_entry);
   }
 }
 
 
 bool
-rpl_parallel::workers_idle(Relay_log_info *rli)
+rpl_parallel::workers_idle()
 {
-  mysql_mutex_assert_owner(&rli->data_lock);
-  return !rli->last_inuse_relaylog ||
-    rli->last_inuse_relaylog->queued_count ==
-    rli->last_inuse_relaylog->dequeued_count;
+  struct rpl_parallel_entry *e;
+  uint32 i, max_i;
+
+  max_i= domain_hash.records;
+  for (i= 0; i < max_i; ++i)
+  {
+    bool active;
+    e= (struct rpl_parallel_entry *)my_hash_element(&domain_hash, i);
+    mysql_mutex_lock(&e->LOCK_parallel_entry);
+    active= e->current_sub_id > e->last_committed_sub_id;
+    mysql_mutex_unlock(&e->LOCK_parallel_entry);
+    if (active)
+      break;
+  }
+  return (i == max_i);
 }
 
 
@@ -2822,7 +2721,7 @@ rpl_parallel::do_event(rpl_group_info *serial_rgi, Log_event *ev,
       else
       {
         DBUG_ASSERT(rli->gtid_skip_flag == GTID_SKIP_TRANSACTION);
-        if (typ == XID_EVENT ||
+        if (typ == XID_EVENT || typ == XA_PREPARE_LOG_EVENT ||
             (typ == QUERY_EVENT &&  // COMMIT/ROLLBACK are never compressed
              (((Query_log_event *)ev)->is_commit() ||
               ((Query_log_event *)ev)->is_rollback())))
@@ -2833,10 +2732,11 @@ rpl_parallel::do_event(rpl_group_info *serial_rgi, Log_event *ev,
     }
   }
 
+  Gtid_log_event *gtid_ev= NULL;
   if (typ == GTID_EVENT)
   {
     rpl_gtid gtid;
-    Gtid_log_event *gtid_ev= static_cast<Gtid_log_event *>(ev);
+    gtid_ev= static_cast<Gtid_log_event *>(ev);
     uint32 domain_id= (rli->mi->using_gtid == Master_info::USE_GTID_NO ||
                        rli->mi->parallel_mode <= SLAVE_PARALLEL_MINIMAL ?
                        0 : gtid_ev->domain_id);
@@ -2875,8 +2775,7 @@ rpl_parallel::do_event(rpl_group_info *serial_rgi, Log_event *ev,
     instead re-use a thread that we queued for previously.
   */
   cur_thread=
-    e->choose_thread(serial_rgi, &did_enter_cond, &old_stage,
-                     typ != GTID_EVENT);
+    e->choose_thread(serial_rgi, &did_enter_cond, &old_stage, gtid_ev);
   if (!cur_thread)
   {
     /* This means we were killed. The error is already signalled. */
@@ -2894,7 +2793,6 @@ rpl_parallel::do_event(rpl_group_info *serial_rgi, Log_event *ev,
 
   if (typ == GTID_EVENT)
   {
-    Gtid_log_event *gtid_ev= static_cast<Gtid_log_event *>(ev);
     bool new_gco;
     enum_slave_parallel_mode mode= rli->mi->parallel_mode;
     uchar gtid_flags= gtid_ev->flags2;
@@ -2937,12 +2835,7 @@ rpl_parallel::do_event(rpl_group_info *serial_rgi, Log_event *ev,
 
       if (mode <= SLAVE_PARALLEL_MINIMAL ||
           !(gtid_flags & Gtid_log_event::FL_GROUP_COMMIT_ID) ||
-          e->last_commit_id != gtid_ev->commit_id ||
-          /*
-            MULTI_BATCH is also set when the current gtid even being a member
-            of a commit group is flagged as DDL which disallows parallel.
-          */
-          (gtid_flags & Gtid_log_event::FL_DDL))
+          e->last_commit_id != gtid_ev->commit_id)
         flags|= group_commit_orderer::MULTI_BATCH;
       /* Make sure we do not attempt to run DDL in parallel speculatively. */
       if (gtid_flags & Gtid_log_event::FL_DDL)
