@@ -1,5 +1,5 @@
 /* Copyright (c) 2011, 2012, Oracle and/or its affiliates.
-   Copyright (c) 2011, 2014, SkySQL Ab.
+   Copyright (c) 2011, 2021, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,7 +27,6 @@
 
 #ifdef __WIN__
 #include <crtdbg.h>
-#include <direct.h>
 #define SIGNAL_FMT "exception 0x%x"
 #else
 #define SIGNAL_FMT "signal %d"
@@ -50,9 +49,6 @@
 static volatile sig_atomic_t segfaulted= 0;
 extern ulong max_used_connections;
 extern volatile sig_atomic_t calling_initgroups;
-#ifdef HAVE_NPTL
-extern volatile sig_atomic_t ld_assume_kernel_is_set;
-#endif
 
 extern const char *optimizer_switch_names[];
 
@@ -63,32 +59,37 @@ static inline void output_core_info()
   char buff[PATH_MAX];
   ssize_t len;
   int fd;
-  if ((len= readlink("/proc/self/cwd", buff, sizeof(buff))) >= 0)
+  if ((len= readlink("/proc/self/cwd", buff, sizeof(buff)-1)) >= 0)
   {
+    buff[len]= 0;
     my_safe_printf_stderr("Writing a core file...\nWorking directory at %.*s\n",
                           (int) len, buff);
   }
-  if ((fd= open("/proc/self/limits", O_RDONLY, MYF(0))) >= 0)
+#ifdef __FreeBSD__
+  if ((fd= my_open("/proc/curproc/rlimit", O_RDONLY, MYF(0))) >= 0)
+#else
+  if ((fd= my_open("/proc/self/limits", O_RDONLY, MYF(0))) >= 0)
+#endif
   {
     my_safe_printf_stderr("Resource Limits:\n");
-    while ((len= read(fd, (uchar*)buff, sizeof(buff))) > 0)
+    while ((len= my_read(fd, (uchar*)buff, sizeof(buff),  MYF(0))) > 0)
     {
       my_write_stderr(buff, len);
     }
-    close(fd);
+    my_close(fd, MYF(0));
   }
 #ifdef __linux__
-  if ((fd= open("/proc/sys/kernel/core_pattern", O_RDONLY, MYF(0))) >= 0)
+  if ((fd= my_open("/proc/sys/kernel/core_pattern", O_RDONLY, MYF(0))) >= 0)
   {
-    len= read(fd, (uchar*)buff, sizeof(buff));
+    len= my_read(fd, (uchar*)buff, sizeof(buff),  MYF(0));
     my_safe_printf_stderr("Core pattern: %.*s\n", (int) len, buff);
-    close(fd);
+    my_close(fd, MYF(0));
   }
-  if ((fd= open("/proc/version", O_RDONLY)) >= 0)
+  if ((fd= my_open("/proc/version", O_RDONLY, MYF(0))) >= 0)
   {
-    len= read(fd, (uchar*)buff, sizeof(buff));
+    len= my_read(fd, (uchar*)buff, sizeof(buff),  MYF(0));
     my_safe_printf_stderr("Kernel version: %.*s\n", (int) len, buff);
-    close(fd);
+    my_close(fd, MYF(0));
   }
 #endif
 #elif defined(__APPLE__) || defined(__FreeBSD__)
@@ -102,14 +103,11 @@ static inline void output_core_info()
   {
     my_safe_printf_stderr("Kernel version: %.*s\n", (int) len, buff);
   }
-#elif defined(HAVE_GETCWD)
+#else
   char buff[80];
-
-  if (getcwd(buff, sizeof(buff)))
-  {
-    my_safe_printf_stderr("Writing a core file at %.*s\n", (int) sizeof(buff), buff);
-    fflush(stderr);
-  }
+  my_getwd(buff, sizeof(buff), 0);
+  my_safe_printf_stderr("Writing a core file at %s\n", buff);
+  fflush(stderr);
 #endif
 }
 
@@ -170,8 +168,11 @@ extern "C" sig_handler handle_fatal_signal(int sig)
   my_safe_printf_stderr("[ERROR] mysqld got " SIGNAL_FMT " ;\n",sig);
 
   my_safe_printf_stderr("%s",
-                        "Sorry, we probably made a mistake, and this is a bug.\n\n"
-                        "Your assistance in bug reporting will enable us to fix this for the next release.\n"
+    "This could be because you hit a bug. It is also possible that this binary\n"
+    "or one of the libraries it was linked against is corrupt, improperly built,\n"
+    "or misconfigured. This error can also be caused by malfunctioning hardware.\n\n");
+
+  my_safe_printf_stderr("%s",
                         "To report this bug, see https://mariadb.com/kb/en/reporting-bugs\n\n");
 
   my_safe_printf_stderr("%s",
@@ -184,35 +185,36 @@ extern "C" sig_handler handle_fatal_signal(int sig)
 		        server_version, SOURCE_REVISION);
 
   if (dflt_key_cache)
-    my_safe_printf_stderr("key_buffer_size=%lu\n",
-                          (ulong) dflt_key_cache->key_cache_mem_size);
+    my_safe_printf_stderr("key_buffer_size=%zu\n",
+                          dflt_key_cache->key_cache_mem_size);
 
-  my_safe_printf_stderr("read_buffer_size=%ld\n",
-                        (long) global_system_variables.read_buff_size);
+  my_safe_printf_stderr("read_buffer_size=%lu\n",
+                        global_system_variables.read_buff_size);
 
   my_safe_printf_stderr("max_used_connections=%lu\n",
-                        (ulong) max_used_connections);
+                        max_used_connections);
 
   if (thread_scheduler)
-    my_safe_printf_stderr("max_threads=%u\n",
-                          (uint) thread_scheduler->max_threads +
-                          (uint) extra_max_connections);
+    my_safe_printf_stderr("max_threads=%lu\n",
+                          thread_scheduler->max_threads +
+                          extra_max_connections);
 
   my_safe_printf_stderr("thread_count=%u\n", THD_count::value());
 
   if (dflt_key_cache && thread_scheduler)
   {
+    size_t used_mem=
+        (dflt_key_cache->key_cache_mem_size +
+         (global_system_variables.read_buff_size +
+          (size_t) global_system_variables.sortbuff_size) *
+             (thread_scheduler->max_threads + extra_max_connections) +
+         (max_connections + extra_max_connections) * sizeof(THD)) / 1024;
+
     my_safe_printf_stderr("It is possible that mysqld could use up to \n"
                           "key_buffer_size + "
                           "(read_buffer_size + sort_buffer_size)*max_threads = "
-                          "%lu K  bytes of memory\n",
-                          (ulong)
-                          (dflt_key_cache->key_cache_mem_size +
-                           (global_system_variables.read_buff_size +
-                            global_system_variables.sortbuff_size) *
-                           (thread_scheduler->max_threads + extra_max_connections) +
-                           (max_connections + extra_max_connections) *
-                           sizeof(THD)) / 1024);
+                          "%zu K  bytes of memory\n", used_mem);
+
     my_safe_printf_stderr("%s",
                           "Hope that's ok; if not, decrease some variables in "
                           "the equation.\n\n");
@@ -303,7 +305,7 @@ extern "C" sig_handler handle_fatal_signal(int sig)
   }
   my_safe_printf_stderr("%s",
     "The manual page at "
-    "https://mariadb.com/kb/en/how-to-produce-a-full-stack-trace-for-mariadbd/ contains\n"
+    "https://mariadb.com/kb/en/how-to-produce-a-full-stack-trace-for-mysqld/ contains\n"
     "information that should help you find out what is causing the crash.\n");
 
 #endif /* HAVE_STACKTRACE */
@@ -322,21 +324,6 @@ extern "C" sig_handler handle_fatal_signal(int sig)
   }
 #endif
 
-#ifdef HAVE_NPTL
-  if (thd_lib_detected == THD_LIB_LT && !ld_assume_kernel_is_set)
-  {
-    my_safe_printf_stderr("%s",
-      "You are running a statically-linked LinuxThreads binary on an NPTL\n"
-      "system. This can result in crashes on some distributions due to "
-      "LT/NPTL conflicts.\n"
-      "You should either build a dynamically-linked binary, "
-      "or force LinuxThreads\n"
-      "to be used with the LD_ASSUME_KERNEL environment variable.\n"
-      "Please consult the documentation for your distribution "
-      "on how to do that.\n");
-  }
-#endif
-
   if (locked_in_memory)
   {
     my_safe_printf_stderr("%s", "\n"
