@@ -1361,7 +1361,7 @@ struct fil_system_t {
 
 private:
   bool m_initialised;
-#ifdef UNIV_LINUX
+#ifdef __linux__
   /** available block devices that reside on non-rotational storage */
   std::vector<dev_t> ssd;
 public:
@@ -1469,9 +1469,10 @@ template<bool have_reference> inline void fil_space_t::flush()
   }
   else if (have_reference)
     flush_low();
-  else if (!(acquire_low() & STOPPING))
+  else
   {
-    flush_low();
+    if (!(acquire_low() & (STOPPING | CLOSING)))
+      flush_low();
     release();
   }
 }
